from os.path import abspath, dirname
import sys

import numpy as np

# Add parent directory to beginning of path variable
sys.path.insert(0, dirname(dirname(abspath(__file__))))
import qpsphere  # noqa: E402


def test_average_sphere_crop():
    radius = 1.01
    hcenter = 2 * radius
    hquarter = 2 * np.sqrt(radius**2 - 1)
    data = np.zeros((5, 5), dtype=float)
    data[2, 2] = 7
    pquarter = 7 * hquarter / hcenter  # ideal sphere
    data[1, 2] = data[3, 2] = data[2, 1] = data[2, 3] = pquarter

    avg, crop = qpsphere.edgefit.average_sphere(image=data,
                                                center=(2, 2),
                                                radius=radius,
                                                weighted=True,
                                                ret_crop=True)
    assert np.all((data == 0) == (crop == 0))
    assert np.allclose(crop[crop != 0], avg)


def test_average_sphere_even():
    radius = 1.01
    hcenter = 2 * radius
    # hquarter is located at the center of the adjacent pixels
    hquarter = 2 * np.sqrt(radius**2 - 2 * .5**2)
    data = np.zeros((4, 4), dtype=float)
    pquarter = 7 * hquarter / hcenter  # ideal sphere
    data[1, 1] = data[1, 2] = data[2, 1] = data[2, 2] = pquarter

    avg = qpsphere.edgefit.average_sphere(image=data,
                                          center=(1.5, 1.5),
                                          radius=radius,
                                          weighted=True,
                                          ret_crop=False)
    assert np.allclose(avg, 7 / (2 * radius))


def test_average_sphere_ideal():
    radius = 1.01
    hcenter = 2 * radius
    hquarter = 2 * np.sqrt(radius**2 - 1)
    data = np.zeros((5, 5), dtype=float)
    data[2, 2] = 7
    pquarter = 7 * hquarter / hcenter  # ideal sphere
    data[1, 2] = data[3, 2] = data[2, 1] = data[2, 3] = pquarter

    avg1 = qpsphere.edgefit.average_sphere(image=data,
                                           center=(2, 2),
                                           radius=radius,
                                           weighted=True,
                                           ret_crop=False)
    avg2 = qpsphere.edgefit.average_sphere(image=data,
                                           center=(2, 2),
                                           radius=radius,
                                           weighted=False,
                                           ret_crop=False)
    assert np.allclose(avg1, avg2), "expected an ideal sphere"
    assert np.allclose(avg1, 7 / (2 * radius)), "expected an ideal sphere"


def test_average_sphere_weighted():
    radius = 1.01
    data = np.zeros((5, 5), dtype=float)
    data[2, 2] = 7
    pquarter = 1  # not an ideal sphere
    data[1, 2] = data[3, 2] = data[2, 1] = data[2, 3] = pquarter

    avg1 = qpsphere.edgefit.average_sphere(image=data,
                                           center=(2, 2),
                                           radius=radius,
                                           weighted=True,
                                           ret_crop=False)
    avg2 = qpsphere.edgefit.average_sphere(image=data,
                                           center=(2, 2),
                                           radius=radius,
                                           weighted=False,
                                           ret_crop=False)
    assert not np.allclose(avg1, avg2), "expected no ideal sphere"
    exact = 7 / (2 * radius)
    assert abs(avg1 - exact) < (avg2 - exact), "weight reduces edge artifacts"


def test_contour_canny_basic():
    size = 21
    cx = 10
    cy = 10
    radius = 7
    data = np.zeros((size, size), dtype=float)
    x = np.arange(size).reshape(-1, 1)
    y = np.arange(size).reshape(1, -1)
    r = np.sqrt((x - cx)**2 + (y - cy)**2)
    data[r <= radius] = 1
    # perform edge detection
    edge = qpsphere.edgefit.contour_canny(image=data,
                                          radius=radius * .9)
    # this might differ from implementation to implementation
    assert np.sum(edge) == 56
    # the edge fully contains `data`
    assert np.allclose(r[edge].max(), 7.2801098892805181)


def test_circle_fit():
    data = np.zeros((7, 7), dtype=bool)
    data[3, 5] = data[1, 3] = data[3, 1] = data[5, 3] = True
    (cx, cy), r, dev = qpsphere.edgefit.circle_fit(data, ret_dev=True)
    assert np.allclose(cx, 3)
    assert np.allclose(cy, 3)
    assert np.allclose(r, 2)
    assert np.allclose(dev, 0, rtol=0, atol=3e-8)


if __name__ == "__main__":
    # Run all tests
    loc = locals()
    for key in list(loc.keys()):
        if key.startswith("test_") and hasattr(loc[key], "__call__"):
            loc[key]()



from .projection import projection


available = {"projection": projection,
             }


def simulate(radius=5, sphere_index=1.339, medium_index=1.333,
             wavelength=550e-9, pixel_size=1e-7, model="projection",
             grid_size=(80, 80), center=(40, 40)):
    """Simulate scattering at a sphere

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the object
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    model: str
        Sphere model to use (see `available`)
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set
    """
    model = available[model]
    qpi = model(radius=radius,
                sphere_index=sphere_index,
                medium_index=medium_index,
                wavelength=wavelength,
                pixel_size=pixel_size,
                grid_size=grid_size,
                center=center)
    return qpi



from ._version import version as __version__  # noqa: F401
from . import edgefit  # noqa: F401
from .models import simulate  # noqa: F401


def analyze(qpi, n0=None, r0=None, method="edge", model="projection",
            ret_center=False, edgekw={}, imagekw={}):
    """Determine refractive index and radius of a spherical object

    Parameters
    ----------
    qpi: QPImage
        Quantitative phase image information
    n0: float
        Approximate refractive index of the phase object,
        estimated with the edge-detection approach if not given
    r0: float
        Approximate radius of the sphere [m]
    method: str
        The method used to determine the refractive index
        can either be "edge" (determine the radius from the
        edge detected in the phase image) or "image" (perform
        a 2D phase image fit).
    model: str
        The light-scattering model used by `method`. If
        `method` is "edge", only "projection" is allowed.
        If `method` is "image", `model` can be one of
        "mie", "projection", "rytov", or "rytov-sc".

    ret_center: bool
        Return the center coordinate of the sphere

    Returns
    -------
    n: float
        Computed refractive index
    r: float
        Computed radius [m]
    center: tuple of floats
        Center position of the sphere [px], only returned
        if `ret_center` is `True`
    """
    if method == "edge":
        if model != "projection":
            raise ValueError("`method='edge'` requires `model='projection'`!")
        if r0 is None:
            raise ValueError("`method='edge'` requires estimate of `r0`!")
        return edgefit.analyze(qpi=qpi,
                               r0=r0,
                               edgekw=edgekw,
                               ret_center=ret_center,
                               ret_edge=False,
                               )
    else:
        raise NotImplementedError("2D phase image fit not yet available!")



"""Refractive index determination for a single cell

This example illustrates how qpsphere can be used to determine
the radius and the refractive index of a spherical cell.
The hologram of the myeloid leukemia cell (HL60) on the left was
recorded using digital holographic microscopy (DHM).
In the quantitative phase image on the right, the detected cell
contour (white) and the subsequent circle fit (red) as well as the
resulting average radius and refractive index of the the cell
are shown. The setup used for recording this data is described in [1]_,
which also contains a description of the basic steps to determine
the position and radius of the cell and to subsequently compute
the average refractive index from the experimental phase data.
The experimental data is loaded and background-corrected using
:py:mod:`qpimage`.


.. [1] M. Schürmann, J. Scholze, P. Müller, C.J. Chan, A.E. Ekpenyong,
   K. Chalut, J. Guck, *Refractive index measurements of single,
   spherical cells using digital holographic microscopy* in Biophysical
   Methods in Cell Biology (ed. E.K. Paluch) 2015,
   `DOI:10.1016/bs.mcb.2014.10.016
   <https://dx.doi.org/10.1016/bs.mcb.2014.10.016>`_
"""
import matplotlib
import matplotlib.pylab as plt
import numpy as np
import qpimage
import qpsphere

# load the experimental data
edata = np.load("./data/hologram_cell.npy.npz")

# create QPImage instance
qpi = qpimage.QPImage(data=edata["data"],
                      bg_data=edata["bg_data"],
                      which_data="hologram",
                      meta_data={"wavelength": 633e-9,
                                 "pixel size": 0.107e-6,
                                 "medium index": 1.335
                                 }
                      )

# background correction
qpi.compute_bg(which_data=["amplitude", "phase"],
               fit_offset="fit",
               fit_profile="ramp",
               border_px=5,
               )

# determine radius and refractive index, guess the cell radius: 10µm
n, r, (cx, cy), edge = qpsphere.edgefit.analyze(qpi=qpi,
                                                r0=10e-6,
                                                ret_center=True,
                                                ret_edge=True)

# plot results
fig = plt.figure(figsize=(8, 4))
matplotlib.rcParams["image.interpolation"] = "bicubic"
holkw = {"cmap": "gray",
         "vmin": 0,
         "vmax": 200}
# hologram image
ax1 = plt.subplot(121, title="cell hologram")
map1 = ax1.imshow(edata["data"].T, **holkw)
plt.colorbar(map1, ax=ax1, fraction=.048, pad=0.04)
# phase image
ax2 = plt.subplot(122, title="phase image [rad]")
map2 = ax2.imshow(qpi.pha.T)
# edge
edgeplot = np.ma.masked_where(edge == 0, edge)
ax2.imshow(edgeplot.T, cmap="gray_r", interpolation="none")
# fitted circle center
plt.plot(cx, cy, "xr", alpha=.5)
# fitted circle perimeter
circle = plt.Circle((cx, cy), r/qpi["pixel size"],
                    color='r', fill=False, ls="dashed", lw=2, alpha=.5)
ax2.add_artist(circle)
# fitting results as text
info = "n={:.4F}\nr={:.2f}µm".format(n, r*1e6)
ax2.text(.8, .8, info, color="w", fontsize="13", verticalalignment="top")
plt.colorbar(map2, ax=ax2, fraction=.048, pad=0.04)
# disable axes
[ax.axis("off") for ax in [ax1, ax2]]

plt.tight_layout()
plt.show()



#!/usr/bin/env python
"""
Determine package version for git repositories.

Each time this file is imported it checks if the ".git" folder is
present and if so, obtains the version from the git history using
`git describe`. This information is then stored in the file
`_version_save.py` which is not versioned by git, but distributed
along with e.g. pypi.
"""
from __future__ import print_function

# Put the entire script into a `True` statement and add the hint
# `pragma: no cover` to ignore code coverage here.
if True:  # pragma: no cover
    import imp
    import os
    from os.path import join, abspath, dirname
    import subprocess
    import sys
    import time
    import traceback
    import warnings

    def git_describe():
        """
        Returns a string describing the version returned by the
        command `git describe --tags HEAD`.
        If it is not possible to determine the correct version,
        then an empty string is returned.
        """
        # make sure we are in a directory that belongs to the correct
        # repository.
        ourdir = dirname(abspath(__file__))

        def _minimal_ext_cmd(cmd):
            # construct minimal environment
            env = {}
            for k in ['SYSTEMROOT', 'PATH']:
                v = os.environ.get(k)
                if v is not None:
                    env[k] = v
            # LANGUAGE is used on win32
            env['LANGUAGE'] = 'C'
            env['LANG'] = 'C'
            env['LC_ALL'] = 'C'
            cmd = subprocess.Popen(cmd, stdout=subprocess.PIPE, env=env)
            out = cmd.communicate()[0]
            return out

        # change directory
        olddir = abspath(os.curdir)
        os.chdir(ourdir)

        try:
            out = _minimal_ext_cmd(['git', 'describe', '--tags', 'HEAD'])
            git_revision = out.strip().decode('ascii')
        except OSError:
            git_revision = ""

        # go back to original directory
        os.chdir(olddir)

        return git_revision

    def load_version(versionfile):
        """ load version from version_save.py
        """
        longversion = ""
        try:
            _version_save = imp.load_source("_version_save", versionfile)
            longversion = _version_save.longversion
        except BaseException:
            try:
                from ._version_save import longversion
            except BaseException:
                try:
                    from _version_save import longversion
                except BaseException:
                    pass

        return longversion

    def save_version(version, versionfile):
        """ save version to version_save.py
        """
        data = "#!/usr/bin/env python\n" \
            + "# This file was created automatically\n" \
            + "longversion = '{VERSION}'\n"
        try:
            with open(versionfile, "w") as fd:
                fd.write(data.format(VERSION=version))
        except BaseException:
            msg = "Could not write package version to {}.".format(versionfile)
            warnings.warn(msg)

    versionfile = join(dirname(abspath(__file__)), "_version_save.py")

    # Determine the accurate version
    longversion = ""

    # 1. git describe
    try:
        # Get the version using `git describe`
        longversion = git_describe()
    except BaseException:
        pass

    # 2. previously created version file
    if longversion == "":
        # Either this is this is not a git repository or we are in the
        # wrong git repository.
        # Get the version from the previously generated `_version_save.py`
        longversion = load_version(versionfile)

    # 3. last resort: date
    if longversion == "":
        print("Could not determine version. Reason:")
        print(traceback.format_exc())
        ctime = os.stat(__file__)[8]
        longversion = time.strftime("%Y.%m.%d-%H-%M-%S", time.gmtime(ctime))
        print("Using creation time as version: {}".format(longversion))

    if not hasattr(sys, 'frozen'):
        # Save the version to `_version_save.py` to allow distribution using
        # `python setup.py sdist`.
        # This is only done if the program is not frozen (with e.g.
        # pyinstaller),
        if longversion != load_version(versionfile):
            save_version(longversion, versionfile)

    # PEP 440-conform development version:
    version = ".dev".join(longversion.split("-")[:2])



""" Geometrical optics (optical path length)
"""
from __future__ import division, print_function, unicode_literals
import numpy as np
import scipy.special as sps

from ..common_methods import average_radial
from ..common_variables import types_indexable


def sphere(
    radius=5,
    n_object=1.339,
    n_medium=1.333,
    distance=0,
    extent=20,
    wavelength=5,
    offset_x=0,
    offset_y=0,
    interpolate=3,
    average=False
):
    """
    Computes the optical path length through a dielectric sphere.

    Parameters
    ----------
    radius : float
        Radius of the cylinder in vacuum wavelengths.
    n_object : float
        Refractive index of the cylinder.
    n_medium : float
        Refractive index of the surrounding medium.
    distance : float
        Distance lD from the detector to the center of the cylinder
        in vacuum wavelengths. If `distance<radius`, the field will be
        automatically refocused to match the behavior of a microscope.
        Does not have any effect here
    extent : float or tuple
        Size of detector in wavelengths. If a float is given, then the
        detector will have a square shape. A tuple is interpreted as the
        rectangular (x,y) size of the detector. 
    wavelength : float
        Resolution of detector in pixels per vacuum wavelength.
    offset_x : float
        Offset in x direction in vacuum wavelengths
    offset_y : float
        Offset in y direction in vacuum wavelengths
    interpolate : int
        If not zero: Instead of computing the entire field, only compute
        the radial field with a sampling that is by a factor of `interpolate`
        higher than the required data and interpolate the 2D field from there.
        This is faster than computing the redundant full field (`interpolate=0`).

    Returns
    -------
    field : 2d ndarray
        The scattered, background corrected field
    """
    if isinstance(extent, types_indexable):
        x = np.linspace(-extent[0] / 2, extent[0] / 2,
                        int(np.round(wavelength * extent[0])), endpoint=True)
        y = np.linspace(-extent[1] / 2, extent[1] / 2,
                        int(np.round(wavelength * extent[1])), endpoint=True)
        cx, cy = np.meshgrid(x, y)
    else:
        x = np.linspace(-extent / 2, extent / 2,
                        int(np.round(wavelength * extent)))
        cy, cx = np.meshgrid(x, x)

    cx += offset_x
    cy += offset_y

    z = np.zeros_like(cx)
    radcoord = radius**2 - cx**2 - cy**2
    radcoordg0 = radcoord > 0
    z[radcoordg0] = 2 * np.sqrt(radcoord[radcoordg0])

    # dn is normalized to one wavelength
    dn = (n_object - n_medium) * z

    if average:
        dn = average_radial(dn, center=(
            offset_x * wavelength, offset_y * wavelength))

    return np.exp(1j * 2 * np.pi * dn).transpose()


def sphere_radial(
    radius=5,
    n_object=1.339,
    n_medium=1.333,
    distance=0,
    radial_extent=20,
    wavelength=5,
    oversample=4,
    average=False
):
    """ Compute radial projection of sphere in a very complicated way

    Parameters
    ----------
    radius : float
        Radius of the cylinder in vacuum wavelengths.
    n_object : float
        Refractive index of the cylinder.
    n_medium : float
        Refractive index of the surrounding medium.
    distance : float
        Distance lD from the detector to the center of the cylinder
        in vacuum wavelengths. If `distance<radius`, the field will be
        automatically refocused to match the behavior of a microscope.
    radial_extent : float or tuple
        Radial size of detector in wavelengths.
    wavelength : float
        Resolution of detector in pixels per vacuum wavelength.
    oversample : int
        Oversampling (zero-padding) to prevent offset in results

    Returns
    -------
    field : 1d ndarray
        The scattered, background corrected radial field
    """
    size = radial_extent * wavelength
    assert int(size) == size
    r = np.linspace(0, radial_extent, size, endpoint=True)
    y = np.zeros_like(r)
    where_disc = r <= radius
    obj2phase = (n_object - n_medium) * 2 * np.pi
    y[where_disc] = 2 * np.sqrt(radius**2 - r[where_disc]**2) * obj2phase

    return np.exp(1j * y)



"""Canny edge detection for QPI analysis of spheres"""
import warnings

import lmfit
import numpy as np
from skimage import feature


class EdgeDetectionError(BaseException):
    pass


class EdgeDetectionWarning(Warning):
    pass


def analyze(qpi, r0, edgekw={}, ret_center=False, ret_edge=False):
    """Determine Refractive index and radius using Canny edge detection

    Compute the refractive index of a spherical phase object by
    detection of an edge in the phase image, a subsequent circle
    fit to the edge, and finally a weighted average over the phase
    image assuming a parabolic phase profile.

    Parameters
    ----------
    qpi: QPImage
        Quantitative phase image information
    r0: float
        Approximate radius of the sphere [m]
    edgekw: dict
        Additional keyword arguments for `contour_canny`
    ret_center: bool
        Return the center coordinate of the sphere

    Returns
    -------
    n: float
        Computed refractive index
    r: float
        Computed radius [m]
    center: tuple of floats
        Center position of the sphere [px], only returned
        if `ret_center` is `True`
    """
    nmed = qpi["medium index"]
    px_m = qpi["pixel size"]
    wl_m = qpi["wavelength"]
    px_wl = px_m / wl_m

    phase = qpi.pha
    # determine edge
    edge = contour_canny(image=phase,
                         radius=r0 / px_m,
                         verbose=False,
                         **edgekw,
                         )
    # fit circle to edge
    center, r = circle_fit(edge)
    # compute average phase density
    avg_phase = average_sphere(phase, center, r)
    # convert phase average to refractive index
    n = nmed + avg_phase / (2 * np.pi * px_wl)
    # convert radius from pixels to meters
    ret = [n, r * px_m]
    if ret_center:
        ret.append(center)
    if ret_edge:
        ret.append(edge)
    return ret


def average_sphere(image, center, radius, weighted=True, ret_crop=False):
    """Compute weighted phase number from a 2D phase image of a sphere

    Parameters
    ----------
    image: 2d ndarray
        Quantitative phase image of a sphere
    center: tuble (x,y)
        Center of the sphere in `image` in ndarray coordinates
    radius: float
        Radius of the sphere in pixels
    weighted: bool
        If `True`, return average phase density weighted with the
        height profile obtained from the radius, otherwise return
        simple average phase density. Weighting gives data points
        at the center of the sphere more weight than those points
        at the boundary of the sphere, avoiding edge artifacts.
    ret_crop: bool
        Return the cropped image.

    Returns
    -------
    average: float
        The average phase value of the sphere from which the refractive
        index can be computed
    cropped_image: 2d ndarray
        Returned if `ret_crop` is True
    """
    sx, sy = image.shape
    x = np.arange(sx).reshape(-1, 1)
    y = np.arange(sy).reshape(1, -1)
    discsq = ((x - center[0])**2 + (y - center[1])**2)
    root = radius**2 - discsq
    # height of the cell for each x and y
    h = 2 * np.sqrt(root * (root > 0))
    # compute phase density
    rho = np.zeros(image.shape)
    hbin = h != 0
    # phase density [rad/px]
    rho[hbin] = image[hbin] / h[hbin]
    if weighted:
        # compute weighted average
        average = np.sum(rho * h) / np.sum(h)
    else:
        # compute simple average
        average = np.sum(rho) / np.sum(hbin)

    ret = average
    if ret_crop:
        ret = (ret, rho)
    return ret


def circle_fit(edge, ret_dev=False):
    """Fit a circle to a boolean edge image

    Parameters
    ----------
    edge: 2d boolean ndarray
        Edge image
    ret_dev: bool
        Return the average deviation of the distance from contour to
        center of the fitted circle.

    Returns
    -------
    center, radius: tuple of floats, float
        Coordinates of the circle. If `ret_dev` is True, then the
        average deviation from the circle is also returned.
    """
    sx, sy = edge.shape
    x = np.linspace(0, sx, sx, endpoint=False).reshape(-1, 1)
    y = np.linspace(0, sy, sy, endpoint=False).reshape(1, -1)
    params = lmfit.Parameters()
    # initial parameters
    sum_edge = np.sum(edge)
    params.add("cx", np.sum(x * edge) / sum_edge)
    params.add("cy", np.sum(y * edge) / sum_edge)
    # data
    xedge, yedge = np.where(edge)
    # minimize
    out = lmfit.minimize(circle_residual, params, args=(xedge, yedge))
    center = (out.params["cx"].value, out.params["cy"].value)
    radii = circle_radii(out.params, xedge, yedge)
    radius = np.mean(radii)

    ret = [center, radius]
    if ret_dev:
        dev = np.average(np.abs(radii - radius))
        ret.append(dev)
    return ret


def circle_radii(params, xedge, yedge):
    cx = params["cx"].value
    cy = params["cy"].value
    radii = np.sqrt((cx - xedge)**2 + (cy - yedge)**2)
    return radii


def circle_residual(params, xedge, yedge):
    radii = circle_radii(params, xedge, yedge)
    return radii - np.mean(radii)


def contour_canny(image, radius, mult_coarse=.40, mult_fine=.1,
                  clip_rmin=.9, clip_rmax=1.1, maxiter=20,
                  verbose=True):
    """Heuristic Canny edge detection for circular objects

    Two Canny-based edge detections with different filter sizes are
    performed to find the outmost contour of an object in a phase image
    while keeping artifacts at a minimum.

    Parameters
    ----------
    image: 2d ndarray
        Image containing an approximately spherically symmetric object
    radius: float
        The approximate object radius in pixels (required for filtering)
    mult_coarse: float
        The coarse edge detection has a filter size of
        ``sigma = mult_coarse * radius``
    mult_fine: float
        The fine edge detection has a filter size of
        ``sigma = mult_fine * radius``
    clip_rmin: float
        Removes edge points that are closer than `clip_rmin` times the
        average radial edge position from the center of the image.
    clip_rmax: float
        Removes edge points that are further than `clip_rmin` times the
        average radial edge position from the center of the image.
    maxiter: int
        Maximum number iterations for coarse edge detection, see Notes
    verbose: bool
        If set to `True`, issues EdgeDetectionWarning where applicable

    Returns
    -------
    edge : 2d boolean ndarray
        The detected edge positions of the object.

    Notes
    -----
    If no edge is found using the filter size defined by `mult_coarse`,
    then the coarse filter size is reduced by a factor of 2 until an
    edge is found or until `maxiter` is reached.

    The edge found using the filter size defined by `mult_fine` is
    heuristically filtered (parts at the center and at the edge of the
    image are removed). This heuristic filtering assumes that the
    circular object is centered in the image.

    See Also
    --------
    skimage.feature.canny: Canny edge detection algorithm used
    """
    image = (image - image.min()) / (image.max() - image.min())

    if radius > image.shape[0] / 2:
        msg = "`radius` in pixels exceeds image size: {}".format(radius)
        raise ValueError(msg)
    # 1. Perform a coarse Canny edge detection. If the edge found is empty,
    # the coarse filter size is reduced by a factor of 2.
    for ii in range(maxiter):
        fact_coarse = .5**ii
        sigma_coarse = radius * mult_coarse * fact_coarse
        edge_coarse = feature.canny(image=image,
                                    sigma=sigma_coarse)
        if np.sum(edge_coarse) != 0:
            break
    else:
        msg = "Could not find edge! Try to reducing `mult_coarse` " \
              + "or increasing `maxiter`."
        raise EdgeDetectionError(msg)

    fact_fine = .7**ii

    if fact_fine != 1 and verbose:
        msg = "The keyword argument `mult_coarse` is too large. " \
              + "If errors occur, adjust `mult_fine` as well.\n" \
              + "Given `mult_coarse`: {}\n".format(mult_coarse) \
              + "New `mult_coarse`: {}\n".format(mult_coarse * fact_coarse) \
              + "Given `mult_fine`: {}\n".format(mult_fine) \
              + "New `mult_fine`: {}".format(mult_fine * fact_fine)
        warnings.warn(msg, EdgeDetectionWarning)

    # 2. Perform a fine Canny edge detection.
    sigma_fine = radius * mult_fine * fact_fine
    edge_fine = feature.canny(image, sigma_fine)

    # 3. Remove parts from the fine edge
    # Assume that the object is centered.
    sx, sy = image.shape
    x = np.linspace(-sx / 2, sx / 2, sx, endpoint=True).reshape(-1, 1)
    y = np.linspace(-sy / 2, sy / 2, sy, endpoint=True).reshape(1, -1)
    # 3.a. Remove detected edge parts from the corners of the image
    # Radius of this disk is approximately
    ellipse = (x / sx)**2 + (y / sy)**2 < .25
    edge_fine *= ellipse
    edge_coarse *= ellipse

    if np.sum(edge_fine):
        # 3.b Also filter inside of edge
        rad = np.sqrt(x**2 + y**2)
        # Filter coarse edge with `clip_rmin` and `clip_rmax`
        rad_coarse = rad * edge_coarse
        avg_coarse = np.sum(rad_coarse) / np.sum(edge_coarse)
        rad_coarse[rad_coarse < avg_coarse * clip_rmin] = 0
        rad_coarse[rad_coarse > avg_coarse * clip_rmax] = 0
        # Filter inside of fine edge with smallest radius of
        # coarse edge, i.e. `rad_coarse.min()`.
        rad_fine = rad * edge_fine
        if np.sum(rad_coarse):
            edge_fine[rad_fine < rad_coarse[rad_coarse != 0].min()] = 0
        # Filter outside of fine edge with `clip_rmax` twice
        for __ in range(2):
            rad_fine = rad * edge_fine
            avg_fine = np.sum(rad_fine) / np.sum(edge_fine)
            edge_fine[rad_fine > avg_fine * clip_rmax] = 0
    elif np.sum(edge_coarse):
        # No fine edge detected.
        edge_fine = edge_coarse
    else:
        msg = "Could not find edge! Try reducing `mult_coarse` " \
              + "and `mult_fine`."
        raise EdgeDetectionError(msg)
    # make sure there are more than 4 points
    if np.sum(edge_fine) < 4:
        msg = "Detected edge too small! Try increasing `maxiter`, " \
              + "modifying `radius`, or reducing `mult_coarse`."
        raise EdgeDetectionError(msg)
    return edge_fine



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from os.path import dirname, realpath, exists
from setuptools import setup, find_packages
import sys


author = u"Paul Müller"
authors = [author]
description = 'library for analyzing spherical objects in phase imaging'
name = 'qpsphere'
year = "2017"

sys.path.insert(0, realpath(dirname(__file__))+"/"+name)
from _version import version

if __name__ == "__main__":
    setup(
        name=name,
        author=author,
        author_email='dev@craban.de',
        url='https://github.com/RI-imaging/qpsphere',
        version=version,
        packages=find_packages(),
        package_dir={name: name},
        license="MIT",
        description=description,
        long_description=open('README.rst').read() if exists('README.rst') else '',
        install_requires=["lmfit",
                          "nrefocus >= 0.1.5",
                          "NumPy >= 1.9.0",
                          "scikit-image >= 0.11.0",
                          "scipy >= 0.18.0",
                          "qpimage",
                          ],
        setup_requires=['pytest-runner'],
        tests_require=["pytest"],
        python_requires='>=3.5, <4',
        keywords=["digital holographic microscopy",
                  "optics",
                  "quantitative phase imaging",
                  "refractive index",
                  "scattering",
                  ],
        classifiers= [
            'Operating System :: OS Independent',
            'Programming Language :: Python :: 3',
            'Intended Audience :: Science/Research'
                     ],
        platforms=['ALL'],
        )



import numpy as np
import qpimage


def projection(radius=5, sphere_index=1.339, medium_index=1.333,
               wavelength=550e-9, pixel_size=.1, grid_size=(80, 80),
               center=(40, 40)):
    """Optical path difference for a dielectric sphere

    Parameters
    ----------
    radius: float
        Radius of the sphere [m]
    sphere_index: float
        Refractive index of the object
    medium_index: float
        Refractive index of the surrounding medium
    wavelength: float
        Vacuum wavelength of the imaging light [m]
    pixel_size: float
        Pixel size [m]
    grid_size: tuple of floats
        Resulting image size in x and y [px]
    center: tuple of floats
        Center position in image coordinates [px]

    Returns
    -------
    qpi: qpimage.QPImage
        Quantitative phase data set
    """
    # grid
    x = np.arange(grid_size[0]).reshape(-1, 1)
    y = np.arange(grid_size[1]).reshape(1, -1)
    cx, cy = center
    # sphere location
    rpx = radius / pixel_size
    r = rpx**2 - (x - cx)**2 - (y - cy)**2
    # distance
    z = np.zeros_like(r)
    rvalid = r > 0
    z[rvalid] = 2 * np.sqrt(r[rvalid]) * pixel_size
    # phase = delta_n * 2PI * z / wavelength
    phase = (sphere_index - medium_index) * 2 * np.pi * z / wavelength
    meta_data = {"pixel size": pixel_size,
                 "wavelength": wavelength,
                 "medium index": medium_index,
                 "sim center": center,
                 "sim radius": radius,
                 "sim index": sphere_index,
                 }
    qpi = qpimage.QPImage(data=phase, which_data="phase",
                          meta_data=meta_data)
    return qpi
